import os
import csv
import json
import shutil
from tqdm import tqdm
import sys

class Parser:

    def __init__(self, from_path, to_path, new_files_file_name, changed_files_file_name, header_values, include_header):
        self.from_path = from_path
        self.to_path = to_path
        self.header_values = header_values
        self.new_files_file_name = new_files_file_name
        self.changed_files_file_name = changed_files_file_name
        self.include_header = include_header

        self.STATE_FILE_NAME = "state.json"

        # reload(sys)
        # sys.setdefaultencoding('utf8')

        csv.register_dialect('custom', delimiter=',', quoting=csv.QUOTE_MINIMAL, quotechar='"', doublequote=True)

    def parse(self):

        unparsed_new_file_paths = []
        unparsed_changed_file_paths = []

        snapshot = self.get_snapshot()

        # Get all new and updated file paths
        lowest_dirs_size = self.get_current_state()

        total_size = 0
        for directory, size in lowest_dirs_size.iteritems():
            if directory in snapshot:
                if size != snapshot[directory]:
                    unparsed_changed_file_paths.append(directory)
                    total_size += size
            else:
                unparsed_new_file_paths.append(directory)
                total_size += size


        if not unparsed_new_file_paths and not unparsed_changed_file_paths:
            print "No new files"

        else:

            header_json = []
            header_text = ""

            first = True
            for header in self.header_values:
                if '=' not in header:
                    header_json.append(header)
                else:
                    split = header.split('=')
                    header_json.append(split[1])
                    if self.include_header:
                        if first:
                            header_text = split[0]
                        else:
                            header_text = header_text + ',' + split[0]
                        first = False

            try:

                # Parse files
                with tqdm(total=total_size, leave=True, unit='B', unit_scale=True) as bar:
                    for path in unparsed_new_file_paths + unparsed_changed_file_paths:
                        full_to_path = os.path.join(self.to_path, 'files', path)
                        full_from_path = os.path.join(self.from_path, 'files', path)
                        updating = False
                        if os.path.exists(full_to_path):
                            os.rename(full_to_path, full_to_path + "_old")
                            updating = True
                        os.makedirs(full_to_path)

                        for file in os.listdir(full_from_path):
                            # Ignore hidden files
                            if not file.startswith('.'):
                                with open(os.path.join(full_from_path, file), 'r') as json_file:
                                    with open(os.path.join(full_to_path, file + '.csv'), 'w+') as csv_file:
                                        csv_writer = csv.DictWriter(csv_file, header_json, dialect='custom', extrasaction='ignore')
                                        if self.include_header:
                                            csv_file.write(header_text + '\n')
                                        for line in json_file:
                                            bar.update(len(line))
                                            parsed_json = json.loads(line)
                                            parsed_json = self.flatten_json(parsed_json)
                                            csv_writer.writerow(parsed_json)

                        if updating:
                            shutil.rmtree(os.path.join(self.to_path, 'files', path + '_old'))

                self.save_state(lowest_dirs_size)

            except KeyboardInterrupt:
                self.rollback(unparsed_new_file_paths, unparsed_changed_file_paths)
                return

    def rollback(self, unparsed_new_file_paths, unparsed_changed_file_paths):
        for path in unparsed_new_file_paths:
            if os.path.exists(os.path.join(self.to_path, 'files', path)):
                shutil.rmtree(os.path.join(self.to_path, 'files', path))
        for path in unparsed_changed_file_paths:
            if os.path.exists(os.path.join(self.to_path, 'files', path + '_old')):
                shutil.rmtree(os.path.join(self.to_path, 'files', path))
                os.rename(os.path.join(self.to_path, 'files', path + '_old'), path)

    def get_snapshot(self):
        if os.path.isfile(os.path.join(self.to_path, self.STATE_FILE_NAME)):
            with open(os.path.join(self.to_path, self.STATE_FILE_NAME), 'r') as state_file:
                return json.load(state_file)
        return {}

    def get_current_state(self):
        dir_with_size = {}
        for root, dirnames, filenames in os.walk(os.path.join(self.from_path, 'files')):
            # If lowest level of folder structure
            if not dirnames:
                path_size = 0
                for file in filenames:
                    # Dont count size of hidden files
                    if not file.startswith('.'):
                        path_size += os.path.getsize(os.path.join(root, file))

                dir_with_size[root.split('files/')[1]] = path_size

        return dir_with_size

    def save_state(self, state):
        with open(os.path.join(self.to_path, self.STATE_FILE_NAME), 'w+') as state_file:
            # for directory, size in state.iteritems():
            json.dump(state, state_file)

    def flatten_json(self,y):
        out = {}

        def flatten(x, name=''):
            if type(x) is dict:
                for a in x:
                    flatten(x[a], name + a + '.')
            elif type(x) is list:
                i = 0
                for a in x:
                    flatten(a, name + str(i) + '.')
                    i += 1
            else:
                out[name[:-1]] = x
                if isinstance(x, unicode):
                    out[str(name[:-1])] = x.encode('utf-8')
                else:
                    out[str(name[:-1])] = str(x)

        flatten(y)
        return out