import sys
from swagger.swagger_client.apis.exportresource_api import ExportresourceApi
from swagger.swagger_client.rest import ApiException
import json
import os.path
import urllib3
import certifi
from tqdm import tqdm
import shutil
import math
import time
import datetime
from apiexecutor import ApiExecutor

class Exports:

    def __init__(self, config_id, path_prefix):
        self.export_api = ExportresourceApi()
        self.path_prefix = path_prefix
        self.config_id = config_id
        self.executor = ApiExecutor()

        # Time until we need to update file links. In seconds
        self.FILE_LINK_INVALID_TIME = 110
        self.STATE_FILE_NAME = "state.json"

    #  Executes the syncronization and prints progress
    def sync_files(self, new_name, changed_name):
        if not changed_name:
            changed_name = 'CHANGED'
        if not new_name:
            new_name = 'NEW'
        try:
            offset = self.get_offset()
        except ValueError:
            print "state.json is corrupted."
            return

        try:
            export = self.executor.execute_with_retry(self.export_api.get_export_using_get, self.config_id)
        except ApiException as e:
            if e.status == 403:
                print "API Access Denied"
            elif e.status == 404:
                print "Export not found"
            else:
                print "Unexpected Error"
            return None
        except urllib3.exceptions.HTTPError as e:
            print "Could not connect and get export info"
            return None
        results = self.get_results(offset)
        if results is None:
            return

        total_files = 0
        total_size = 0
        for export_result in results.results:
            total_files = total_files + export_result.file_count
            total_size = total_size + export_result.estimated_total_size

        if total_files == 0:
            print "Local files are up to date."
            return

        print "Downloading " + str(total_files) + " files"

        http = urllib3.PoolManager(
            cert_reqs='CERT_REQUIRED', # Force certificate check.
            ca_certs=certifi.where()  # Path to the Certifi bundle.
            )

        is_hourly = export.time_unit == "HOURS"

        file_counter = 0
        updated_paths = []
        new_paths = []
        with tqdm(total=total_size, leave=True, unit='B', unit_scale=True) as bar:
            for export_result in results.results:

                file_name_counter = 0

                if (export_result.file_count > 0):

                    # Results are off wih about 300 milliseconds, adding 400 to get correct date and hours.
                    start_time = export_result.instance_start_time + datetime.timedelta(milliseconds=400)

                    time_path = 'dt=' + start_time.strftime("%Y-%m-%d")
                    if is_hourly:
                        time_path = os.path.join(time_path, 'hh=' + start_time.strftime("%H"))

                    path = os.path.join(self.path_prefix, 'files', time_path)

                    try:
                        updating_data = False
                        if os.path.exists(path):
                            updating_data = True
                            os.rename(path, path + "_old")
                        os.makedirs(path)

                        refresh_links = time.time() + self.FILE_LINK_INVALID_TIME

                        files = self.executor.execute_with_retry(self.export_api.get_files_using_get, self.config_id, export_result.id)

                        for export_file in files._files:
                            file_with_url = self.executor.execute_with_retry(self.export_api.get_file_url_using_get, self.config_id, export_result.id, export_file.id)

                            if file_with_url is None:
                                raise ValueError("file can not be None")

                            response = http.request("GET", file_with_url.content_url, preload_content=False)

                            CHUNK = 16 * 1024

                            file_path = os.path.join(path, 'data' + '-' + str(file_name_counter))
                            with open(file_path, 'w+') as file_on_disk:
                                while True:
                                    chunk = response.read(CHUNK)
                                    if not chunk:
                                        break
                                    file_on_disk.write(chunk)
                                    bar.update(len(chunk))
                            file_counter += 1
                            file_name_counter += 1

                        if updating_data:
                            updated_paths.append(time_path)
                            shutil.rmtree(path + "_old")
                        else:
                            new_paths.append(time_path)

                        self.save_state(export_result.offset)

                    except (urllib3.exceptions.HTTPError) as e:
                        print "Network error occurred, exiting."
                        self.handle_exception(updating_data, path, updated_paths, new_paths, new_name, changed_name)
                        return
                    except ApiException as e:
                        print 'Error occured when connecting to export service, check that your credentials are valid'
                        self.handle_exception(updating_data, path, updated_paths, new_paths, new_name, changed_name)
                        return
                    except (KeyboardInterrupt) as e:
                        self.handle_exception(updating_data, path, updated_paths, new_paths, new_name, changed_name)
                        return
                    except:
                        print "Unexpected error occurred"
                        self.handle_exception(updating_data, path, updated_paths, new_paths, new_name, changed_name)
                        return

        self.write_log(updated_paths, new_paths, new_name, changed_name)
        print str(file_counter) + " files successfully downloaded"

    def handle_exception(self, updating_data, path, updated_paths, new_paths, new_name, changed_name):
        if updating_data:
            if os.path.exists(path + "_old"):
                shutil.rmtree(path)
                os.rename(path + "_old", path)
        elif os.path.exists(path):
            shutil.rmtree(path)

        self.write_log(updated_paths, new_paths, new_name, changed_name)

    def write_log(self, updated_paths, new_paths, new_files_log_name, changed_files_log_name):
        with open(os.path.join(self.path_prefix, changed_files_log_name), 'w+') as file:
            for path in updated_paths:
                file.write(path + '\n')
        with open(os.path.join(self.path_prefix, new_files_log_name), 'w+') as file:
            for path in new_paths:
                file.write(path + '\n')

    # Will print error message and return None if http error or no config found while fetching results
    def get_results(self, offset):
        try:
            return self.executor.execute_with_retry(self.export_api.get_results_using_get, self.config_id, offset=offset)
        except ApiException as e:
            if e.status == 403:
                print "API Access Denied"
            elif e.status == 404:
                print "Export not found"
            else:
                print "Unexpected Error"
            return None
        except urllib3.exceptions.HTTPError as e:
            print "Could not connect and get export results"
            return None

    def get_info_json(self):
        data = self.get_info()
        if data is not None:
            return json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))

    def get_info_readable(self):
        data = self.get_info()

        if data is not None:
            return data['start_time'] + '  -  ' + data['end_time'] + "\nInterval: " + str(data['interval']) + "\nTime Unit: " + data['time_unit'] + "\nRemote data size: " + self.convert_size(data['remote_data_size']) +"\nLocal data size: " + self.convert_size(data['local_data_size']) + "\nUnsynced data size: " + self.convert_size(data['unsynced_data_size'])
        else:
            return ""

    def get_info(self):
        results = self.get_results(0)
        if results is None:
            return None

        try:
            export = self.executor.execute_with_retry(self.export_api.get_export_using_get, self.config_id)
        except ApiException as e:
            if e.status == 403:
                print "API Access Denied"
            elif e.status == 404:
                print "Export not found"
            else:
                print "Unexpected Error"
                print str(e)
            return None
        except urllib3.exceptions.HTTPError as e:
            print "Could not connect and get export info"
            return None
        
        try:
            offset = self.get_offset()
        except ValueError:
            print "state.json is not valid."
            return None

        data = {}

        data['start_time'] = str(export.start_time)
        data['end_time'] = str(export.end_time)
        data['interval'] = export.interval
        data['time_unit'] = export.time_unit

        total_size = 0
        unsynced_size = 0

        # Key: instance_start_time, Value: (data_version, estimated_total_size)
        checked_exports = {}
        for export_result in results.results:
            if checked_exports.get(export_result.instance_start_time, None) is not None:
                if export_result.data_version > checked_exports[export_result.instance_start_time][0]:
                    total_size -= checked_exports[export_result.instance_start_time][1]
                    total_size += export_result.estimated_total_size

                    if export_result.offset > offset:
                        unsynced_size -= checked_exports[export_result.instance_start_time][1]
                        unsynced_size += export_result.estimated_total_size

                    checked_exports[export_result.instance_start_time] = (export_result.data_version, export_result.estimated_total_size)

                    
            else:
                checked_exports[export_result.instance_start_time] = (export_result.data_version, export_result.estimated_total_size)

                total_size = total_size + export_result.estimated_total_size

                if export_result.offset > offset:
                    unsynced_size += export_result.estimated_total_size

        data['remote_data_size'] = total_size
        data['unsynced_data_size'] = unsynced_size
        local_size = 0
        for root, dirs, files, in os.walk(os.path.join(self.path_prefix, 'files')):
            local_size += sum(os.path.getsize(os.path.join(root, name)) for name in files)

        data['local_data_size'] = local_size

        return data

    def convert_size(self, size):
        if (size == 0):
            return '0B'
        size_name = ("B","KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        i = int(math.floor(math.log(size,1000)))
        p = math.pow(1000,i)
        s = round(size/p,2)
        return '%s %s' % (s,size_name[i])

    # will return 0 if no last id
    def get_offset(self):
        if os.path.isfile(os.path.join(self.path_prefix, self.STATE_FILE_NAME)):
            with open(os.path.join(self.path_prefix, self.STATE_FILE_NAME), 'r') as state_file:
                data = json.load(state_file)
                return data["offset"]

        return 0

    def save_state(self, new_state):
        with open(os.path.join(self.path_prefix, self.STATE_FILE_NAME), 'w+') as state_file:
            json.dump({'offset': new_state}, state_file)